@@ -37,13 +37,14 @@ struct _PhocLayerSurface {
     struct wl_listener output_destroy;
     struct wl_listener new_popup;
     struct wl_listener new_subsurface;
-    struct wl_list subsurfaces; // roots_layer_subsurface::link
+    struct wl_list subsurfaces; // phoc_layer_subsurface::link
 
     struct wlr_box geo;
     enum zwlr_layer_shell_v1_layer layer;
+    bool mapped;
 };
 
-PhocLayerSurface *phoc_layer_surface_new (void);
+PhocLayerSurface *phoc_layer_surface_new (struct wlr_layer_surface_v1 *layer_surface);
 void              phoc_layer_surface_unmap (PhocLayerSurface *self);
 const char       *phoc_layer_surface_get_namespace (PhocLayerSurface *self);
 PhocOutput       *phoc_layer_surface_get_output (PhocLayerSurface *self);
