@@ -5,16 +5,15 @@
 #ifndef _POSIX_C_SOURCE
 #define _POSIX_C_SOURCE 200112L
 #endif
-#include <assert.h>
 #include <errno.h>
 #include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 #include <wayland-server-core.h>
-#include <wlr/types/wlr_box.h>
 #include <wlr/types/wlr_surface.h>
 #include <wlr/types/wlr_layer_shell_v1.h>
+#include <wlr/util/box.h>
 #include "cursor.h"
 #include "desktop.h"
 #include "layers.h"
@@ -22,6 +21,8 @@
 #include "seat.h"
 #include "server.h"
 
+#include <glib.h>
+
 #define LAYER_SHELL_LAYER_COUNT 4
 
 static void apply_exclusive(struct wlr_box *usable_area,
@@ -86,8 +87,7 @@ static void apply_exclusive(struct wlr_box *usable_area,
 	}
 }
 
-static void update_cursors(PhocLayerSurface *roots_surface,
-		GSList *seats /* PhocSeat */) {
+static void update_cursors(PhocLayerSurface *layer_surface, GSList *seats /* PhocSeat */) {
 	PhocServer *server = phoc_server_get_default ();
 
 	for (GSList *elem = phoc_input_get_seats (server->input); elem; elem = elem->next) {
@@ -101,7 +101,7 @@ static void update_cursors(PhocLayerSurface *roots_surface,
 			server->desktop,
 			cursor->cursor->x, cursor->cursor->y, &sx, &sy, NULL);
 
-		if (surface == roots_surface->layer_surface->surface) {
+		if (surface == layer_surface->layer_surface->surface) {
 			struct timespec time;
 			if (clock_gettime(CLOCK_MONOTONIC, &time) == 0) {
 				phoc_cursor_update_position(cursor,
@@ -114,17 +114,23 @@ static void update_cursors(PhocLayerSurface *roots_surface,
 	}
 }
 
-static void arrange_layer(struct wlr_output *output,
+static void arrange_layer(PhocOutput *output,
 		GSList *seats /* PhocSeat */,
-		struct wl_list *list /* PhocLayerSurface */,
+		enum zwlr_layer_shell_v1_layer layer,
 		struct wlr_box *usable_area, bool exclusive) {
-	PhocLayerSurface *roots_surface;
+	PhocLayerSurface *layer_surface;
 	struct wlr_box full_area = { 0 };
-	wlr_output_effective_resolution(output,
+
+	g_assert (PHOC_IS_OUTPUT (output));
+	wlr_output_effective_resolution(output->wlr_output,
 			&full_area.width, &full_area.height);
-	wl_list_for_each_reverse(roots_surface, list, link) {
-		struct wlr_layer_surface_v1 *layer = roots_surface->layer_surface;
-		struct wlr_layer_surface_v1_state *state = &layer->current;
+	wl_list_for_each_reverse(layer_surface, &output->layer_surfaces, link) {
+		struct wlr_layer_surface_v1 *wlr_layer_surface = layer_surface->layer_surface;
+		struct wlr_layer_surface_v1_state *state = &wlr_layer_surface->current;
+
+		if (layer_surface->layer != layer)
+		  continue;
+
 		if (exclusive != (state->exclusive_zone > 0)) {
 			continue;
 		}
@@ -183,19 +189,21 @@ static void arrange_layer(struct wlr_output *output,
 		}
 		if (box.width < 0 || box.height < 0) {
 			// TODO: Bubble up a protocol error?
-			wlr_layer_surface_v1_close(layer);
+			wlr_layer_surface_v1_destroy(wlr_layer_surface);
 			continue;
 		}
 
 		// Apply
-		struct wlr_box old_geo = roots_surface->geo;
-		roots_surface->geo = box;
-		if (layer->mapped) {
+		struct wlr_box old_geo = layer_surface->geo;
+		layer_surface->geo = box;
+		if (wlr_layer_surface->mapped) {
 			apply_exclusive(usable_area, state->anchor, state->exclusive_zone,
 					state->margin.top, state->margin.right,
 					state->margin.bottom, state->margin.left);
 		}
-		wlr_layer_surface_v1_configure(layer, box.width, box.height);
+
+		if (box.width != old_geo.width || box.height != old_geo.height)
+			wlr_layer_surface_v1_configure(wlr_layer_surface, box.width, box.height);
 
 		// Having a cursor newly end up over the moved layer will not
 		// automatically send a motion event to the surface. The event needs to
@@ -203,47 +211,35 @@ static void arrange_layer(struct wlr_output *output,
 		// Only update layer surfaces which kept their size (and so buffers) the
 		// same, because those with resized buffers will be handled separately.
 
-		if (roots_surface->geo.x != old_geo.x
-				|| roots_surface->geo.y != old_geo.y) {
-			update_cursors(roots_surface, seats);
+		if (layer_surface->geo.x != old_geo.x
+				|| layer_surface->geo.y != old_geo.y) {
+			update_cursors(layer_surface, seats);
 		}
 	}
 }
 
-struct osk_origin {
-	struct wlr_layer_surface_v1_state state;
-	PhocLayerSurface *surface;
-	enum zwlr_layer_shell_v1_layer layer;
-};
+static PhocLayerSurface *
+find_osk (PhocOutput *output)
+{
+  PhocLayerSurface *layer_surface;
 
-static struct osk_origin find_osk(struct wl_list layers[LAYER_SHELL_LAYER_COUNT]) {
-	struct osk_origin origin = {0};
-	for (unsigned i = 0; i < LAYER_SHELL_LAYER_COUNT; i++) {
-		struct wl_list *list = &layers[i];
-		PhocLayerSurface *roots_surface;
-		wl_list_for_each(roots_surface, list, link) {
-			if (strcmp(roots_surface->layer_surface->namespace, "osk") == 0) {
-				origin.state = roots_surface->layer_surface->current;
-				origin.surface = roots_surface;
-				origin.layer = i;
-				return origin;
-			}
-		}
-	}
-	return origin;
+  wl_list_for_each(layer_surface, &output->layer_surfaces, link) {
+    if (strcmp(layer_surface->layer_surface->namespace, "osk") == 0)
+      return layer_surface;
+  }
+
+  return NULL;
 }
 
 /// Adjusts keyboard properties
-static void change_osk(const struct osk_origin *osk, struct wl_list layers[LAYER_SHELL_LAYER_COUNT], bool force_overlay) {
-	if (force_overlay && osk->layer != ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY) {
-		wl_list_remove(&osk->surface->link);
-		wl_list_insert(&layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY], &osk->surface->link);
-	}
+static void
+change_osk (PhocLayerSurface *osk, struct wl_list layer_surfaces, bool force_overlay)
+{
+  if (force_overlay && osk->layer != ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY)
+    osk->layer = ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY;
 
-	if (!force_overlay && osk->layer != osk->surface->layer_surface->client_pending.layer) {
-		wl_list_remove(&osk->surface->link);
-		wl_list_insert(&layers[osk->surface->layer_surface->client_pending.layer], &osk->surface->link);
-	}
+  if (!force_overlay && osk->layer != osk->layer_surface->pending.layer)
+    osk->layer = osk->layer_surface->pending.layer;
 }
 
 void
@@ -261,25 +257,26 @@ phoc_layer_shell_arrange (PhocOutput *output)
 
   wlr_output_effective_resolution (output->wlr_output, &usable_area.width, &usable_area.height);
 
-  struct osk_origin osk_place = find_osk (output->layers);
-  if (osk_place.surface) {
+  PhocLayerSurface *osk = find_osk (output);
+  if (osk) {
     bool osk_force_overlay = false;
 
     for (GSList *elem = seats; elem; elem = elem->next) {
       PhocSeat *seat = PHOC_SEAT (elem->data);
 
       g_assert (PHOC_IS_SEAT (seat));
-      if (seat->focused_layer && seat->focused_layer->client_pending.layer >= osk_place.surface->layer_surface->client_pending.layer) {
+      if (seat->focused_layer && seat->focused_layer->pending.layer >= osk->layer_surface->pending.layer &&
+          phoc_input_method_relay_is_enabled (&seat->im_relay, seat->focused_layer->surface)) {
         osk_force_overlay = true;
         break;
       }
     }
-    change_osk (&osk_place, output->layers, osk_force_overlay);
+    change_osk (osk, output->layer_surfaces, osk_force_overlay);
   }
 
   // Arrange exclusive surfaces from top->bottom
   for (size_t i = 0; i < G_N_ELEMENTS(layers); ++i)
-    arrange_layer (output->wlr_output, seats, &output->layers[layers[i]], &usable_area, true);
+    arrange_layer (output, seats, layers[i], &usable_area, true);
   output->usable_area = usable_area;
 
   PhocView *view;
@@ -295,7 +292,7 @@ phoc_layer_shell_arrange (PhocOutput *output)
 
   // Arrange non-exlusive surfaces from top->bottom
   for (size_t i = 0; i < G_N_ELEMENTS(layers); ++i)
-    arrange_layer (output->wlr_output, seats, &output->layers[layers[i]], &usable_area, false);
+    arrange_layer (output, seats, layers[i], &usable_area, false);
 }
 
 void
@@ -306,19 +303,39 @@ phoc_layer_shell_update_focus (void)
     ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY,
     ZWLR_LAYER_SHELL_V1_LAYER_TOP,
   };
-  PhocLayerSurface *layer, *topmost = NULL;
+  PhocLayerSurface *layer_surface, *topmost = NULL;
   // Find topmost keyboard interactive layer, if such a layer exists
   // TODO: Make layer surface focus per-output based on cursor position
   PhocOutput *output;
   wl_list_for_each (output, &server->desktop->outputs, link) {
     for (size_t i = 0; i < G_N_ELEMENTS(layers_above_shell); ++i) {
-      if (output->fullscreen_view && !output->force_shell_reveal) {
-        if (layers_above_shell[i] == ZWLR_LAYER_SHELL_V1_LAYER_TOP)
+      wl_list_for_each_reverse (layer_surface, &output->layer_surfaces, link) {
+        if (layer_surface->layer != layers_above_shell[i])
           continue;
+
+        if (layer_surface->layer_surface->current.exclusive_zone <= 0)
+          continue;
+
+        if (layer_surface->layer_surface->current.keyboard_interactive &&
+            layer_surface->layer_surface->mapped) {
+          topmost = layer_surface;
+          break;
+        }
       }
-      wl_list_for_each(layer, &output->layers[layers_above_shell[i]], link) {
-        if (layer->layer_surface->current.keyboard_interactive && layer->layer_surface->mapped) {
-          topmost = layer;
+
+      if (topmost != NULL)
+        break;
+
+      wl_list_for_each (layer_surface, &output->layer_surfaces, link) {
+        if (layer_surface->layer != layers_above_shell[i])
+          continue;
+
+        if (layer_surface->layer_surface->current.exclusive_zone > 0)
+          continue;
+
+        if (layer_surface->layer_surface->current.keyboard_interactive &&
+            layer_surface->layer_surface->mapped) {
+          topmost = layer_surface;
           break;
         }
       }
@@ -335,35 +352,24 @@ phoc_layer_shell_update_focus (void)
   }
 }
 
-static void handle_output_destroy(struct wl_listener *listener, void *data) {
-	PhocLayerSurface *layer =
-		wl_container_of(listener, layer, output_destroy);
-	layer->layer_surface->output = NULL;
-	wl_list_remove(&layer->output_destroy.link);
-	wlr_layer_surface_v1_close(layer->layer_surface);
-}
-
 static void handle_surface_commit(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	PhocLayerSurface *layer =
-		wl_container_of(listener, layer, surface_commit);
-	struct wlr_layer_surface_v1 *layer_surface = layer->layer_surface;
-	struct wlr_output *wlr_output = layer_surface->output;
+	PhocLayerSurface *layer_surface = wl_container_of(listener, layer_surface, surface_commit);
+	struct wlr_layer_surface_v1 *wlr_layer_surface = layer_surface->layer_surface;
+	struct wlr_output *wlr_output = wlr_layer_surface->output;
 	if (wlr_output != NULL) {
 		PhocOutput *output = wlr_output->data;
-		struct wlr_box old_geo = layer->geo;
+		struct wlr_box old_geo = layer_surface->geo;
 
-		bool layer_changed = layer->layer != layer_surface->current.layer;
-		if (layer_changed) {
-			wl_list_remove(&layer->link);
-			wl_list_insert(&output->layers[layer_surface->current.layer],
-				&layer->link);
-			layer->layer = layer_surface->current.layer;
+		bool layer_changed = false;
+		if (wlr_layer_surface->current.committed != 0) {
+			layer_changed = layer_surface->layer != wlr_layer_surface->current.layer;
+
+			layer_surface->layer = wlr_layer_surface->current.layer;
+			phoc_layer_shell_arrange (output);
+			phoc_layer_shell_update_focus ();
 		}
 
-		phoc_layer_shell_arrange (output);
-		phoc_layer_shell_update_focus ();
-
 		// Cursor changes which happen as a consequence of resizing a layer
 		// surface are applied in phoc_layer_shell_arrange. Because the resize happens
 		// before the underlying surface changes, it will only receive a cursor
@@ -371,34 +377,33 @@ static void handle_surface_commit(struct wl_listener *listener, void *data) {
 		// the *new* layer surface.
 		// Another cursor move event is needed when the surface actually
 		// changes.
-		struct wlr_surface *surface = layer_surface->surface;
+		struct wlr_surface *surface = wlr_layer_surface->surface;
 		if (surface->previous.width != surface->current.width ||
 				surface->previous.height != surface->current.height) {
-			update_cursors(layer, phoc_input_get_seats (server->input));
+			update_cursors(layer_surface, phoc_input_get_seats (server->input));
 		}
 
 		bool geo_changed =
-			memcmp(&old_geo, &layer->geo, sizeof(struct wlr_box)) != 0;
+			memcmp(&old_geo, &layer_surface->geo, sizeof(struct wlr_box)) != 0;
 		if (geo_changed || layer_changed) {
-			phoc_output_damage_whole_local_surface(output, layer_surface->surface,
+			phoc_output_damage_whole_local_surface(output, wlr_layer_surface->surface,
 							       old_geo.x, old_geo.y);
-			phoc_output_damage_whole_local_surface(output, layer_surface->surface,
-							       layer->geo.x, layer->geo.y);
+			phoc_output_damage_whole_local_surface(output, wlr_layer_surface->surface,
+							       layer_surface->geo.x, layer_surface->geo.y);
 		} else {
-			phoc_output_damage_from_local_surface(output, layer_surface->surface,
-							      layer->geo.x, layer->geo.y);
+			phoc_output_damage_from_local_surface(output, wlr_layer_surface->surface,
+							      layer_surface->geo.x, layer_surface->geo.y);
 		}
 	}
 }
 
 
 static void handle_destroy(struct wl_listener *listener, void *data) {
-	PhocLayerSurface *layer = wl_container_of(
-			listener, layer, destroy);
-	g_object_unref (layer);
+	PhocLayerSurface *layer_surface = wl_container_of(listener, layer_surface, destroy);
+	g_object_unref (layer_surface);
 }
 
-static void subsurface_destroy(struct roots_layer_subsurface *subsurface) {
+static void subsurface_destroy(struct phoc_layer_subsurface *subsurface) {
 	wl_list_remove(&subsurface->map.link);
 	wl_list_remove(&subsurface->unmap.link);
 	wl_list_remove(&subsurface->destroy.link);
@@ -407,17 +412,17 @@ static void subsurface_destroy(struct roots_layer_subsurface *subsurface) {
 	free(subsurface);
 }
 
-static struct roots_layer_popup *popup_create(struct wlr_xdg_popup *wlr_popup);
-static struct roots_layer_subsurface *layer_subsurface_create(struct wlr_subsurface *wlr_subsurface);
+static struct phoc_layer_popup *popup_create(struct wlr_xdg_popup *wlr_popup);
+static struct phoc_layer_subsurface *layer_subsurface_create(struct wlr_subsurface *wlr_subsurface);
 
-static PhocLayerSurface *popup_get_root_layer(struct roots_layer_popup *popup) {
+static PhocLayerSurface *popup_get_root_layer(struct phoc_layer_popup *popup) {
 	while (popup->parent_type == LAYER_PARENT_POPUP) {
 		popup = popup->parent_popup;
 	}
 	return popup->parent_layer;
 }
 
-static void popup_unconstrain(struct roots_layer_popup *popup) {
+static void popup_unconstrain(struct phoc_layer_popup *popup) {
 	PhocLayerSurface *layer = popup_get_root_layer(popup);
 	struct wlr_xdg_popup *wlr_popup = popup->wlr_popup;
 
@@ -435,11 +440,11 @@ static void popup_unconstrain(struct roots_layer_popup *popup) {
 	wlr_xdg_popup_unconstrain_from_box(wlr_popup, &output_toplevel_sx_box);
 }
 
-static void popup_damage(struct roots_layer_popup *layer_popup, bool whole) {
+static void popup_damage(struct phoc_layer_popup *layer_popup, bool whole) {
 	struct wlr_xdg_popup *popup = layer_popup->wlr_popup;
 	struct wlr_surface *surface = popup->base->surface;
-	int popup_sx = popup->geometry.x - popup->base->geometry.x;
-	int popup_sy = popup->geometry.y - popup->base->geometry.y;
+	int popup_sx = popup->geometry.x - popup->base->current.geometry.x;
+	int popup_sy = popup->geometry.y - popup->base->current.geometry.y;
 	int ox = popup_sx, oy = popup_sy;
 	PhocLayerSurface *layer;
 	while (layer_popup->parent_type == LAYER_PARENT_POPUP) {
@@ -464,20 +469,20 @@ static void popup_damage(struct roots_layer_popup *layer_popup, bool whole) {
 }
 
 static void popup_new_popup(struct wl_listener *listener, void *data) {
-	struct roots_layer_popup *popup =
+	struct phoc_layer_popup *popup =
 		wl_container_of(listener, popup, new_popup);
 	struct wlr_xdg_popup *wlr_popup = data;
-	struct roots_layer_popup *new_popup = popup_create(wlr_popup);
+	struct phoc_layer_popup *new_popup = popup_create(wlr_popup);
 	new_popup->parent_type = LAYER_PARENT_POPUP;
 	new_popup->parent_popup = popup;
 	popup_unconstrain(new_popup);
 }
 
 static void popup_new_subsurface(struct wl_listener *listener, void *data) {
-	struct roots_layer_popup *popup =
+	struct phoc_layer_popup *popup =
 		wl_container_of(listener, popup, new_subsurface);
 	struct wlr_subsurface *wlr_subsurface = data;
-	struct roots_layer_subsurface *subsurface = layer_subsurface_create(wlr_subsurface);
+	struct phoc_layer_subsurface *subsurface = layer_subsurface_create(wlr_subsurface);
 	subsurface->parent_type = LAYER_PARENT_POPUP;
 	subsurface->parent_popup = popup;
 	wl_list_insert(&popup->subsurfaces, &subsurface->link);
@@ -485,7 +490,7 @@ static void popup_new_subsurface(struct wl_listener *listener, void *data) {
 
 static void popup_handle_map(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	struct roots_layer_popup *popup = wl_container_of(listener, popup, map);
+	struct phoc_layer_popup *popup = wl_container_of(listener, popup, map);
 	PhocLayerSurface *layer = popup_get_root_layer(popup);
 	struct wlr_output *wlr_output = layer->layer_surface->output;
 	if (!wlr_output) {
@@ -493,14 +498,14 @@ static void popup_handle_map(struct wl_listener *listener, void *data) {
 	}
 
 	struct wlr_subsurface *child;
-	wl_list_for_each(child, &popup->wlr_popup->base->surface->subsurfaces_below, parent_link) {
-		struct roots_layer_subsurface *new_subsurface = layer_subsurface_create(child);
+	wl_list_for_each(child, &popup->wlr_popup->base->surface->current.subsurfaces_below, current.link) {
+		struct phoc_layer_subsurface *new_subsurface = layer_subsurface_create(child);
 		new_subsurface->parent_type = LAYER_PARENT_POPUP;
 		new_subsurface->parent_popup = popup;
 		wl_list_insert(&popup->subsurfaces, &new_subsurface->link);
 	}
-	wl_list_for_each(child, &popup->wlr_popup->base->surface->subsurfaces_above, parent_link) {
-		struct roots_layer_subsurface *new_subsurface = layer_subsurface_create(child);
+	wl_list_for_each(child, &popup->wlr_popup->base->surface->current.subsurfaces_above, current.link) {
+		struct phoc_layer_subsurface *new_subsurface = layer_subsurface_create(child);
 		new_subsurface->parent_type = LAYER_PARENT_POPUP;
 		new_subsurface->parent_popup = popup;
 		wl_list_insert(&popup->subsurfaces, &new_subsurface->link);
@@ -515,8 +520,8 @@ static void popup_handle_map(struct wl_listener *listener, void *data) {
 
 static void popup_handle_unmap(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	struct roots_layer_popup *popup = wl_container_of(listener, popup, unmap);
-	struct roots_layer_subsurface *child, *tmp;
+	struct phoc_layer_popup *popup = wl_container_of(listener, popup, unmap);
+	struct phoc_layer_subsurface *child, *tmp;
 	wl_list_for_each_safe(child, tmp, &popup->subsurfaces, link) {
 		subsurface_destroy(child);
 	}
@@ -526,12 +531,12 @@ static void popup_handle_unmap(struct wl_listener *listener, void *data) {
 }
 
 static void popup_handle_commit(struct wl_listener *listener, void *data) {
-	struct roots_layer_popup *popup = wl_container_of(listener, popup, commit);
+	struct phoc_layer_popup *popup = wl_container_of(listener, popup, commit);
 	popup_damage(popup, false);
 }
 
 static void popup_handle_destroy(struct wl_listener *listener, void *data) {
-	struct roots_layer_popup *popup =
+	struct phoc_layer_popup *popup =
 		wl_container_of(listener, popup, destroy);
 
 	wl_list_remove(&popup->map.link);
@@ -542,9 +547,9 @@ static void popup_handle_destroy(struct wl_listener *listener, void *data) {
 	free(popup);
 }
 
-static struct roots_layer_popup *popup_create(struct wlr_xdg_popup *wlr_popup) {
-	struct roots_layer_popup *popup =
-		calloc(1, sizeof(struct roots_layer_popup));
+static struct phoc_layer_popup *popup_create(struct wlr_xdg_popup *wlr_popup) {
+	struct phoc_layer_popup *popup =
+		calloc(1, sizeof(struct phoc_layer_popup));
 	if (popup == NULL) {
 		return NULL;
 	}
@@ -566,16 +571,16 @@ static struct roots_layer_popup *popup_create(struct wlr_xdg_popup *wlr_popup) {
 }
 
 static void handle_new_popup(struct wl_listener *listener, void *data) {
-	PhocLayerSurface *roots_layer_surface =
-		wl_container_of(listener, roots_layer_surface, new_popup);
+	PhocLayerSurface *phoc_layer_surface =
+		wl_container_of(listener, phoc_layer_surface, new_popup);
 	struct wlr_xdg_popup *wlr_popup = data;
-	struct roots_layer_popup *popup = popup_create(wlr_popup);
+	struct phoc_layer_popup *popup = popup_create(wlr_popup);
 	popup->parent_type = LAYER_PARENT_LAYER;
-	popup->parent_layer = roots_layer_surface;
+	popup->parent_layer = phoc_layer_surface;
 	popup_unconstrain(popup);
 }
 
-static PhocLayerSurface *subsurface_get_root_layer(struct roots_layer_subsurface *subsurface) {
+static PhocLayerSurface *subsurface_get_root_layer(struct phoc_layer_subsurface *subsurface) {
 	while (subsurface->parent_type == LAYER_PARENT_SUBSURFACE) {
 		subsurface = subsurface->parent_subsurface;
 	}
@@ -585,7 +590,7 @@ static PhocLayerSurface *subsurface_get_root_layer(struct roots_layer_subsurface
 	return subsurface->parent_layer;
 }
 
-static void subsurface_damage(struct roots_layer_subsurface *subsurface, bool whole) {
+static void subsurface_damage(struct phoc_layer_subsurface *subsurface, bool whole) {
 	PhocLayerSurface *layer = subsurface_get_root_layer(subsurface);
 	PhocOutput *output = phoc_layer_surface_get_output (layer);
 	if (!output) {
@@ -605,11 +610,11 @@ static void subsurface_damage(struct roots_layer_subsurface *subsurface, bool wh
 }
 
 static void subsurface_new_subsurface(struct wl_listener *listener, void *data) {
-	struct roots_layer_subsurface *subsurface =
+	struct phoc_layer_subsurface *subsurface =
 		wl_container_of(listener, subsurface, new_subsurface);
 	struct wlr_subsurface *wlr_subsurface = data;
 
-	struct roots_layer_subsurface *new_subsurface = layer_subsurface_create(wlr_subsurface);
+	struct phoc_layer_subsurface *new_subsurface = layer_subsurface_create(wlr_subsurface);
 	new_subsurface->parent_type = LAYER_PARENT_SUBSURFACE;
 	new_subsurface->parent_subsurface = subsurface;
 	wl_list_insert(&subsurface->subsurfaces, &new_subsurface->link);
@@ -617,17 +622,17 @@ static void subsurface_new_subsurface(struct wl_listener *listener, void *data)
 
 static void subsurface_handle_map(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	struct roots_layer_subsurface *subsurface = wl_container_of(listener, subsurface, map);
+	struct phoc_layer_subsurface *subsurface = wl_container_of(listener, subsurface, map);
 
 	struct wlr_subsurface *child;
-	wl_list_for_each(child, &subsurface->wlr_subsurface->surface->subsurfaces_below, parent_link) {
-		struct roots_layer_subsurface *new_subsurface = layer_subsurface_create(child);
+	wl_list_for_each(child, &subsurface->wlr_subsurface->surface->current.subsurfaces_below, current.link) {
+		struct phoc_layer_subsurface *new_subsurface = layer_subsurface_create(child);
 		new_subsurface->parent_type = LAYER_PARENT_SUBSURFACE;
 		new_subsurface->parent_subsurface = subsurface;
 		wl_list_insert(&subsurface->subsurfaces, &new_subsurface->link);
 	}
-	wl_list_for_each(child, &subsurface->wlr_subsurface->surface->subsurfaces_above, parent_link) {
-		struct roots_layer_subsurface *new_subsurface = layer_subsurface_create(child);
+	wl_list_for_each(child, &subsurface->wlr_subsurface->surface->current.subsurfaces_above, current.link) {
+		struct phoc_layer_subsurface *new_subsurface = layer_subsurface_create(child);
 		new_subsurface->parent_type = LAYER_PARENT_SUBSURFACE;
 		new_subsurface->parent_subsurface = subsurface;
 		wl_list_insert(&subsurface->subsurfaces, &new_subsurface->link);
@@ -642,8 +647,8 @@ static void subsurface_handle_map(struct wl_listener *listener, void *data) {
 
 static void subsurface_handle_unmap(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	struct roots_layer_subsurface *subsurface = wl_container_of(listener, subsurface, unmap);
-	struct roots_layer_subsurface *child, *tmp;
+	struct phoc_layer_subsurface *subsurface = wl_container_of(listener, subsurface, unmap);
+	struct phoc_layer_subsurface *child, *tmp;
 	wl_list_for_each_safe(child, tmp, &subsurface->subsurfaces, link) {
 		subsurface_destroy(child);
 	}
@@ -653,20 +658,20 @@ static void subsurface_handle_unmap(struct wl_listener *listener, void *data) {
 }
 
 static void subsurface_handle_commit(struct wl_listener *listener, void *data) {
-	struct roots_layer_subsurface *subsurface = wl_container_of(listener, subsurface, commit);
+	struct phoc_layer_subsurface *subsurface = wl_container_of(listener, subsurface, commit);
 	subsurface_damage(subsurface, false);
 }
 
 static void subsurface_handle_destroy(struct wl_listener *listener, void *data) {
-	struct roots_layer_subsurface *subsurface =
+	struct phoc_layer_subsurface *subsurface =
 		wl_container_of(listener, subsurface, destroy);
 
 	subsurface_destroy(subsurface);
 }
 
-static struct roots_layer_subsurface *layer_subsurface_create(struct wlr_subsurface *wlr_subsurface) {
-	struct roots_layer_subsurface *subsurface =
-		calloc(1, sizeof(struct roots_layer_subsurface));
+static struct phoc_layer_subsurface *layer_subsurface_create(struct wlr_subsurface *wlr_subsurface) {
+	struct phoc_layer_subsurface *subsurface =
+		calloc(1, sizeof(struct phoc_layer_subsurface));
 	if (subsurface == NULL) {
 		return NULL;
 	}
@@ -688,82 +693,93 @@ static struct roots_layer_subsurface *layer_subsurface_create(struct wlr_subsurf
 }
 
 static void handle_new_subsurface(struct wl_listener *listener, void *data) {
-	PhocLayerSurface *roots_layer_surface =
-		wl_container_of(listener, roots_layer_surface, new_subsurface);
+	PhocLayerSurface *phoc_layer_surface =
+		wl_container_of(listener, phoc_layer_surface, new_subsurface);
 	struct wlr_subsurface *wlr_subsurface = data;
 
-	struct roots_layer_subsurface *subsurface = layer_subsurface_create(wlr_subsurface);
+	struct phoc_layer_subsurface *subsurface = layer_subsurface_create(wlr_subsurface);
 	subsurface->parent_type = LAYER_PARENT_LAYER;
-	subsurface->parent_layer = roots_layer_surface;
-	wl_list_insert(&roots_layer_surface->subsurfaces, &subsurface->link);
+	subsurface->parent_layer = phoc_layer_surface;
+	wl_list_insert(&phoc_layer_surface->subsurfaces, &subsurface->link);
 }
 
 static void handle_map(struct wl_listener *listener, void *data) {
-	struct wlr_layer_surface_v1 *layer_surface = data;
-	PhocLayerSurface *layer = PHOC_LAYER_SURFACE (layer_surface->data);
-	PhocOutput *output = phoc_layer_surface_get_output (layer);
+	struct wlr_layer_surface_v1 *wlr_layer_surface = data;
+	PhocLayerSurface *layer_surface = PHOC_LAYER_SURFACE (wlr_layer_surface->data);
+	PhocOutput *output = phoc_layer_surface_get_output (layer_surface);
 	if (!output) {
 		return;
 	}
 
+	layer_surface->mapped = true;
+
 	struct wlr_subsurface *subsurface;
-	wl_list_for_each(subsurface, &layer_surface->surface->subsurfaces_below, parent_link) {
-		struct roots_layer_subsurface *roots_subsurface = layer_subsurface_create(subsurface);
-		roots_subsurface->parent_type = LAYER_PARENT_LAYER;
-		roots_subsurface->parent_layer = layer;
-		wl_list_insert(&layer->subsurfaces, &roots_subsurface->link);
+	wl_list_for_each(subsurface, &wlr_layer_surface->surface->current.subsurfaces_below, current.link) {
+		struct phoc_layer_subsurface *phoc_subsurface = layer_subsurface_create(subsurface);
+		phoc_subsurface->parent_type = LAYER_PARENT_LAYER;
+		phoc_subsurface->parent_layer = layer_surface;
+		wl_list_insert(&layer_surface->subsurfaces, &phoc_subsurface->link);
 	}
-	wl_list_for_each(subsurface, &layer_surface->surface->subsurfaces_above, parent_link) {
-		struct roots_layer_subsurface *roots_subsurface = layer_subsurface_create(subsurface);
-		roots_subsurface->parent_type = LAYER_PARENT_LAYER;
-		roots_subsurface->parent_layer = layer;
-		wl_list_insert(&layer->subsurfaces, &roots_subsurface->link);
+	wl_list_for_each(subsurface, &wlr_layer_surface->surface->current.subsurfaces_above, current.link) {
+		struct phoc_layer_subsurface *phoc_subsurface = layer_subsurface_create(subsurface);
+		phoc_subsurface->parent_type = LAYER_PARENT_LAYER;
+		phoc_subsurface->parent_layer = layer_surface;
+		wl_list_insert(&layer_surface->subsurfaces, &phoc_subsurface->link);
 	}
 
-	layer->new_subsurface.notify = handle_new_subsurface;
-	wl_signal_add(&layer_surface->surface->events.new_subsurface, &layer->new_subsurface);
+	layer_surface->new_subsurface.notify = handle_new_subsurface;
+	wl_signal_add(&wlr_layer_surface->surface->events.new_subsurface, &layer_surface->new_subsurface);
 
 	phoc_output_damage_whole_local_surface(output,
-					       layer_surface->surface, layer->geo.x,
-					       layer->geo.y);
-	wlr_surface_send_enter(layer_surface->surface, output->wlr_output);
+					       wlr_layer_surface->surface, layer_surface->geo.x,
+					       layer_surface->geo.y);
+	wlr_surface_send_enter(wlr_layer_surface->surface, output->wlr_output);
+
+	phoc_layer_shell_arrange (output);
+	phoc_layer_shell_update_focus ();
 }
 
 static void handle_unmap(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	PhocLayerSurface *layer = wl_container_of(
-			listener, layer, unmap);
+	PhocLayerSurface *layer_surface = wl_container_of(listener, layer_surface, unmap);
+	PhocOutput *output = phoc_layer_surface_get_output (layer_surface);
 
-	struct roots_layer_subsurface *subsurface, *tmp;
-	wl_list_for_each_safe(subsurface, tmp, &layer->subsurfaces, link) {
+	layer_surface->mapped = false;
+
+	struct phoc_layer_subsurface *subsurface, *tmp;
+	wl_list_for_each_safe(subsurface, tmp, &layer_surface->subsurfaces, link) {
 		subsurface_destroy(subsurface);
 	}
-	wl_list_remove(&layer->new_subsurface.link);
+	wl_list_remove(&layer_surface->new_subsurface.link);
 
-	phoc_layer_surface_unmap (layer);
+	phoc_layer_surface_unmap (layer_surface);
 	phoc_input_update_cursor_focus(server->input);
+
+	if (output)
+		phoc_layer_shell_arrange (output);
+	phoc_layer_shell_update_focus ();
 }
 
 void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 	PhocServer *server = phoc_server_get_default ();
-	struct wlr_layer_surface_v1 *layer_surface = data;
-	PhocDesktop *desktop =
-		wl_container_of(listener, desktop, layer_shell_surface);
+	struct wlr_layer_surface_v1 *wlr_layer_surface = data;
+	PhocDesktop *desktop = wl_container_of(listener, desktop, layer_shell_surface);
+
 	g_debug ("new layer surface: namespace %s layer %d anchor %d "
 			"size %dx%d margin %d,%d,%d,%d",
-		layer_surface->namespace, layer_surface->client_pending.layer,
-		layer_surface->client_pending.anchor,
-		layer_surface->client_pending.desired_width,
-		layer_surface->client_pending.desired_height,
-		layer_surface->client_pending.margin.top,
-		layer_surface->client_pending.margin.right,
-		layer_surface->client_pending.margin.bottom,
-		layer_surface->client_pending.margin.left);
+		wlr_layer_surface->namespace, wlr_layer_surface->pending.layer,
+		wlr_layer_surface->pending.anchor,
+		wlr_layer_surface->pending.desired_width,
+		wlr_layer_surface->pending.desired_height,
+		wlr_layer_surface->pending.margin.top,
+		wlr_layer_surface->pending.margin.right,
+		wlr_layer_surface->pending.margin.bottom,
+		wlr_layer_surface->pending.margin.left);
 
-	if (!layer_surface->output) {
+	if (!wlr_layer_surface->output) {
 		PhocInput *input = server->input;
 		PhocSeat *seat = phoc_input_get_last_active_seat(input);
-		assert(seat); // Technically speaking we should handle this case
+		g_assert (PHOC_IS_SEAT (seat)); // Technically speaking we should handle this case
 		PhocCursor *cursor = phoc_seat_get_cursor(seat);
 		struct wlr_output *output =
 			wlr_output_layout_output_at(desktop->layout,
@@ -776,50 +792,39 @@ void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 			output = wlr_output_layout_get_center_output(desktop->layout);
 		}
 		if (output) {
-			layer_surface->output = output;
+			wlr_layer_surface->output = output;
 		} else {
-			wlr_layer_surface_v1_close(layer_surface);
+			wlr_layer_surface_v1_destroy(wlr_layer_surface);
 			return;
 		}
 	}
 
-	PhocLayerSurface *roots_surface = phoc_layer_surface_new ();
-	if (!roots_surface) {
-		return;
-	}
 
-	wl_list_init(&roots_surface->subsurfaces);
+	PhocLayerSurface *layer_surface = phoc_layer_surface_new (wlr_layer_surface);
 
-	roots_surface->surface_commit.notify = handle_surface_commit;
-	wl_signal_add(&layer_surface->surface->events.commit,
-		&roots_surface->surface_commit);
+	layer_surface->surface_commit.notify = handle_surface_commit;
+	wl_signal_add(&wlr_layer_surface->surface->events.commit,
+		&layer_surface->surface_commit);
 
-	roots_surface->output_destroy.notify = handle_output_destroy;
-	wl_signal_add(&layer_surface->output->events.destroy,
-		&roots_surface->output_destroy);
+	layer_surface->destroy.notify = handle_destroy;
+	wl_signal_add(&wlr_layer_surface->events.destroy, &layer_surface->destroy);
+	layer_surface->map.notify = handle_map;
+	wl_signal_add(&wlr_layer_surface->events.map, &layer_surface->map);
+	layer_surface->unmap.notify = handle_unmap;
+	wl_signal_add(&wlr_layer_surface->events.unmap, &layer_surface->unmap);
+	layer_surface->new_popup.notify = handle_new_popup;
+	wl_signal_add(&wlr_layer_surface->events.new_popup, &layer_surface->new_popup);
 
-	roots_surface->destroy.notify = handle_destroy;
-	wl_signal_add(&layer_surface->events.destroy, &roots_surface->destroy);
-	roots_surface->map.notify = handle_map;
-	wl_signal_add(&layer_surface->events.map, &roots_surface->map);
-	roots_surface->unmap.notify = handle_unmap;
-	wl_signal_add(&layer_surface->events.unmap, &roots_surface->unmap);
-	roots_surface->new_popup.notify = handle_new_popup;
-	wl_signal_add(&layer_surface->events.new_popup, &roots_surface->new_popup);
+	PhocOutput *output = wlr_layer_surface->output->data;
+	wl_list_insert(&output->layer_surfaces, &layer_surface->link);
 
-	roots_surface->layer_surface = layer_surface;
-	layer_surface->data = roots_surface;
-
-	PhocOutput *output = layer_surface->output->data;
-	wl_list_insert(&output->layers[layer_surface->client_pending.layer], &roots_surface->link);
-
-	// Temporarily set the layer's current state to client_pending
+	// Temporarily set the layer's current state to pending
 	// So that we can easily arrange it
-	struct wlr_layer_surface_v1_state old_state = layer_surface->current;
-	layer_surface->current = layer_surface->client_pending;
+	struct wlr_layer_surface_v1_state old_state = wlr_layer_surface->current;
+	wlr_layer_surface->current = wlr_layer_surface->pending;
 
 	phoc_layer_shell_arrange (output);
 	phoc_layer_shell_update_focus ();
 
-	layer_surface->current = old_state;
+	wlr_layer_surface->current = old_state;
 }
