@@ -3,7 +3,6 @@
 #include "config.h"
 
 #define _POSIX_C_SOURCE 200809L
-#include <assert.h>
 #include <stdbool.h>
 #include <stdlib.h>
 #include <time.h>
@@ -14,17 +13,36 @@
 #include <wlr/types/wlr_output_power_management_v1.h>
 #include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/util/region.h>
+
+#include "anim/animatable.h"
 #include "settings.h"
 #include "layers.h"
+#include "layer-shell-effects.h"
 #include "output.h"
+#include "output-shield.h"
 #include "render.h"
+#include "render-private.h"
 #include "seat.h"
 #include "server.h"
 #include "utils.h"
 #include "xwayland-surface.h"
 
+static void phoc_output_initable_iface_init (GInitableIface *iface);
 
-G_DEFINE_TYPE (PhocOutput, phoc_output, G_TYPE_OBJECT);
+typedef struct _PhocOutputPrivate {
+  PhocOutputShield *shield;
+
+  GSList *frame_callbacks;
+  gint    frame_callback_next_id;
+  gint64  last_frame_us;
+
+  gboolean shell_revealed;
+  gboolean force_shell_reveal;
+} PhocOutputPrivate;
+
+G_DEFINE_TYPE_WITH_CODE (PhocOutput, phoc_output, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (PhocOutput)
+                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, phoc_output_initable_iface_init));
 
 enum {
   PROP_0,
@@ -41,6 +59,15 @@ enum {
 static guint signals[N_SIGNALS] = { 0 };
 
 
+typedef struct {
+  PhocAnimatable    *animatable;
+  PhocFrameCallback  callback;
+  gpointer           user_data;
+  GDestroyNotify     notify;
+  guint              id;
+} PhocOutputFrameCallbackInfo;
+
+
 typedef struct {
   PhocSurfaceIterator  user_iterator;
   void                *user_data;
@@ -52,6 +79,15 @@ typedef struct {
 } PhocOutputSurfaceIteratorData;
 
 
+static void
+phoc_output_frame_callback_info_free (PhocOutputFrameCallbackInfo *cb_info)
+{
+  if (cb_info->notify && cb_info->user_data)
+    cb_info->notify (cb_info->user_data);
+  g_free (cb_info);
+}
+
+
 static bool
 get_surface_box (PhocOutputSurfaceIteratorData *data,
                  struct wlr_surface *surface, int sx, int sy,
@@ -85,7 +121,7 @@ get_surface_box (PhocOutputSurfaceIteratorData *data,
 
   struct wlr_box rotated_box;
 
-  wlr_box_rotated_bounds (&rotated_box, &box, data->rotation);
+  phoc_utils_rotated_bounds (&rotated_box, &box, data->rotation);
 
   struct wlr_box output_box = {0};
 
@@ -95,7 +131,7 @@ get_surface_box (PhocOutputSurfaceIteratorData *data,
 
   struct wlr_box intersection;
 
-  return wlr_box_intersection (&intersection, &output_box, &rotated_box);
+  return wlr_box_intersection (&intersection, &output_box, &box);
 }
 
 
@@ -146,15 +182,23 @@ phoc_output_get_property (GObject    *object,
 static void
 phoc_output_init (PhocOutput *self)
 {
+  PhocOutputPrivate *priv = phoc_output_get_instance_private(self);
+
+  priv->frame_callback_next_id = 1;
+  priv->last_frame_us = g_get_monotonic_time ();
+  priv->shield = phoc_output_shield_new (self);
+
+  self->debug_touch_points = NULL;
+  wl_list_init (&self->layer_surfaces);
 }
 
 PhocOutput *
-phoc_output_new (PhocDesktop *desktop, struct wlr_output *wlr_output)
+phoc_output_new (PhocDesktop *desktop, struct wlr_output *wlr_output, GError **error)
 {
-  return g_object_new (PHOC_TYPE_OUTPUT,
-                       "desktop", desktop,
-                       "wlr-output", wlr_output,
-                       NULL);
+  return g_initable_new (PHOC_TYPE_OUTPUT, NULL, error,
+                         "desktop", desktop,
+                         "wlr-output", wlr_output,
+                         NULL);
 }
 
 static void
@@ -202,10 +246,30 @@ phoc_output_damage_handle_frame (struct wl_listener *listener,
                                  void               *data)
 {
   PhocOutput *self = wl_container_of (listener, self, damage_frame);
+  PhocOutputPrivate *priv = phoc_output_get_instance_private (self);
   PhocServer *server = phoc_server_get_default ();
   PhocRenderer *renderer = phoc_server_get_renderer (server);
 
+  GSList *l = priv->frame_callbacks;
+  while (l != NULL) {
+    GSList *next = l->next;
+    PhocOutputFrameCallbackInfo *cb_info = l->data;
+    gboolean ret;
+
+    ret = cb_info->callback(cb_info->animatable, priv->last_frame_us, cb_info->user_data);
+    if (ret == G_SOURCE_REMOVE) {
+      phoc_output_frame_callback_info_free (cb_info);
+      priv->frame_callbacks = g_slist_delete_link (priv->frame_callbacks, l);
+    }
+    l = next;
+  }
+  priv->last_frame_us = g_get_monotonic_time ();
+
   phoc_renderer_render_output (renderer, self);
+
+  /* Want frame clock ticking as long as we have frame callbacks */
+  if (priv->frame_callbacks)
+    wlr_output_schedule_frame(self->wlr_output);
 }
 
 static void
@@ -240,6 +304,47 @@ phoc_output_handle_commit (struct wl_listener *listener, void *data)
   phoc_layer_shell_arrange (self);
 }
 
+static float
+phoc_output_compute_scale (struct wlr_output *output)
+{
+  int32_t width = 0, height = 0;
+
+  if (!output->phys_width || !output->phys_height) {
+    g_message ("Output '%s' has invalid physical size, "
+               "using default scale", output->name);
+    return 1;
+  }
+
+  // Use the pending mode if any
+  if (output->pending.committed & WLR_OUTPUT_STATE_MODE) {
+    switch (output->pending.mode_type) {
+    case WLR_OUTPUT_STATE_MODE_FIXED:
+      width = output->pending.mode->width;
+      height = output->pending.mode->height;
+      break;
+    case WLR_OUTPUT_STATE_MODE_CUSTOM:
+      width = output->pending.custom_mode.width;
+      height = output->pending.custom_mode.height;
+      break;
+    default:
+      break;
+    }
+  // Fall back to current mode
+  } else if (output->current_mode) {
+    width = output->current_mode->width;
+    height = output->current_mode->height;
+  }
+
+  if (!width || !height) {
+    g_message ("No valid mode set for output '%s', "
+               "using default scale", output->name);
+    return 1;
+  }
+
+  return phoc_utils_compute_scale (output->phys_width, output->phys_height,
+                                   width, height);
+}
+
 static void
 phoc_output_set_mode (struct wlr_output *output, PhocOutputConfig *oc)
 {
@@ -272,14 +377,18 @@ phoc_output_set_mode (struct wlr_output *output, PhocOutputConfig *oc)
   }
 }
 
-static void
-phoc_output_constructed (GObject *object)
+
+static gboolean
+phoc_output_initable_init (GInitable    *initable,
+                           GCancellable *cancellable,
+                           GError      **error)
 {
-  PhocOutput *self = PHOC_OUTPUT (object);
+  PhocOutput *self = PHOC_OUTPUT (initable);
   PhocServer *server = phoc_server_get_default ();
+  PhocRenderer *renderer = phoc_server_get_renderer (server);
   PhocInput *input = server->input;
 
-  assert (server->desktop);
+  g_assert (PHOC_IS_DESKTOP (server->desktop));
 
   PhocConfig *config = self->desktop->config;
 
@@ -295,9 +404,16 @@ phoc_output_constructed (GObject *object)
   self->wlr_output->data = self;
   wl_list_insert (&self->desktop->outputs, &self->link);
 
-  self->damage = wlr_output_damage_create (self->wlr_output);
+  if (!wlr_output_init_render (self->wlr_output,
+                               phoc_renderer_get_wlr_allocator (renderer),
+                               phoc_renderer_get_wlr_renderer (renderer))) {
+    g_set_error (error,
+                 G_FILE_ERROR, G_FILE_ERROR_FAILED,
+		 "Could not create renderer");
+    return FALSE;
+  }
 
-  self->debug_touch_points = NULL;
+  self->damage = wlr_output_damage_create (self->wlr_output);
 
   self->output_destroy.notify = phoc_output_handle_destroy;
   wl_signal_add (&self->wlr_output->events.destroy, &self->output_destroy);
@@ -313,9 +429,6 @@ phoc_output_constructed (GObject *object)
   self->damage_destroy.notify = phoc_output_damage_handle_destroy;
   wl_signal_add (&self->damage->events.destroy, &self->damage_destroy);
 
-  for (size_t i = 0; i < G_N_ELEMENTS (self->layers); ++i)
-    wl_list_init (&self->layers[i]);
-
   PhocOutputConfig *output_config = phoc_config_get_output (config, self->wlr_output);
 
   struct wlr_output_mode *preferred_mode =
@@ -338,7 +451,12 @@ phoc_output_constructed (GObject *object)
         wlr_output_set_mode (self->wlr_output, preferred_mode);
       }
 
-      wlr_output_set_scale (self->wlr_output, output_config->scale);
+      if (!output_config->scale) {
+        wlr_output_set_scale (self->wlr_output,
+                              phoc_output_compute_scale (self->wlr_output));
+      } else {
+        wlr_output_set_scale (self->wlr_output, output_config->scale);
+      }
       wlr_output_set_transform (self->wlr_output, output_config->transform);
       wlr_output_layout_add (self->desktop->layout, self->wlr_output,
                              output_config->x, output_config->y);
@@ -349,6 +467,8 @@ phoc_output_constructed (GObject *object)
     if (preferred_mode != NULL) {
       wlr_output_set_mode (self->wlr_output, preferred_mode);
     }
+    wlr_output_set_scale (self->wlr_output,
+                          phoc_output_compute_scale (self->wlr_output));
     wlr_output_enable (self->wlr_output, true);
     wlr_output_layout_add_auto (self->desktop->layout, self->wlr_output);
   }
@@ -368,14 +488,14 @@ phoc_output_constructed (GObject *object)
 
   update_output_manager_config (self->desktop);
 
-  G_OBJECT_CLASS (phoc_output_parent_class)->constructed (object);
-
+  return TRUE;
 }
 
 static void
 phoc_output_finalize (GObject *object)
 {
   PhocOutput *self = PHOC_OUTPUT (object);
+  PhocOutputPrivate *priv = phoc_output_get_instance_private (self);
 
   wl_list_remove (&self->link);
   wl_list_remove (&self->enable.link);
@@ -383,15 +503,24 @@ phoc_output_finalize (GObject *object)
   wl_list_remove (&self->commit.link);
   wl_list_remove (&self->output_destroy.link);
   g_clear_list (&self->debug_touch_points, g_free);
+  /* Remove all frame callbacks, this will also free associzted user data */
+  g_clear_slist (&priv->frame_callbacks,
+                 (GDestroyNotify)phoc_output_frame_callback_info_free);
 
-  for (size_t i = 0; i < G_N_ELEMENTS (self->layers); ++i)
-    wl_list_remove (&self->layers[i]);
+  wl_list_init (&self->layer_surfaces);
 
+  g_clear_object (&priv->shield);
   g_clear_object (&self->desktop);
 
   G_OBJECT_CLASS (phoc_output_parent_class)->finalize (object);
 }
 
+static void
+phoc_output_initable_iface_init (GInitableIface *iface)
+{
+  iface->init = phoc_output_initable_init;
+}
+
 static void
 phoc_output_class_init (PhocOutputClass *klass)
 {
@@ -400,7 +529,6 @@ phoc_output_class_init (PhocOutputClass *klass)
   object_class->set_property = phoc_output_set_property;
   object_class->get_property = phoc_output_get_property;
 
-  object_class->constructed = phoc_output_constructed;
   object_class->finalize = phoc_output_finalize;
 
   props[PROP_DESKTOP] =
@@ -427,7 +555,9 @@ phoc_output_class_init (PhocOutputClass *klass)
 
 static void
 phoc_output_for_each_surface_iterator (struct wlr_surface *surface,
-                                       int sx, int sy, void *_data)
+                                       int                 sx,
+                                       int                 sy,
+                                       void               *_data)
 {
   PhocOutputSurfaceIteratorData *data = _data;
 
@@ -442,11 +572,24 @@ phoc_output_for_each_surface_iterator (struct wlr_surface *surface,
                        data->scale, data->user_data);
 }
 
+/**
+ * phoc_output_surface_for_each_surface:
+ * @self: the output
+ * @surface: The wlr_surface
+ * @ox: x position in output coordinates
+ * @oy: y position in output coordinates
+ * @iterator: (scope call): The callback invoked on each iteration
+ * @user_data: Callback user data
+ *
+ * Iterate over surfaces in a `wlr_surface`s surface tree.
+ */
 void
-phoc_output_surface_for_each_surface (PhocOutput *self, struct wlr_surface
-                                      *surface, double ox, double oy,
-                                      PhocSurfaceIterator iterator,
-                                      void *user_data)
+phoc_output_surface_for_each_surface (PhocOutput          *self,
+                                      struct wlr_surface  *surface,
+                                      double               ox,
+                                      double               oy,
+                                      PhocSurfaceIterator  iterator,
+                                      void                *user_data)
 {
   PhocOutputSurfaceIteratorData data = {
     .user_iterator = iterator,
@@ -464,12 +607,24 @@ phoc_output_surface_for_each_surface (PhocOutput *self, struct wlr_surface
                                 phoc_output_for_each_surface_iterator, &data);
 }
 
+/**
+ * phoc_output_xdg_surface_for_each_surface:
+ * @self: the output
+ * @xdg_surface: The wlr_xdg_surface
+ * @ox: x position in output coordinates
+ * @oy: y position in output coordinates
+ * @iterator: (scope call): The callback invoked on each iteration
+ * @user_data: Callback user data
+ *
+ * Iterate over surfaces in a `wlr_xdg_surface`s surface tree.
+ */
 void
-phoc_output_xdg_surface_for_each_surface (PhocOutput *self, struct
-                                          wlr_xdg_surface *xdg_surface, double
-                                          ox, double oy,
-                                          PhocSurfaceIterator
-                                          iterator, void *user_data)
+phoc_output_xdg_surface_for_each_surface (PhocOutput             *self,
+                                          struct wlr_xdg_surface *xdg_surface,
+                                          double                  ox,
+                                          double                  oy,
+                                          PhocSurfaceIterator     iterator,
+                                          void                   *user_data)
 {
   PhocOutputSurfaceIteratorData data = {
     .user_iterator = iterator,
@@ -487,10 +642,20 @@ phoc_output_xdg_surface_for_each_surface (PhocOutput *self, struct
                                     phoc_output_for_each_surface_iterator, &data);
 }
 
+/**
+ * phoc_output_view_for_each_surface:
+ * @self: the output
+ * @view: The [type@View]
+ * @iterator: (scope call): The callback invoked on each iteration
+ * @user_data: Callback user data
+ *
+ * Iterate over surfaces in a [type@View]s surface tree.
+ */
 void
-phoc_output_view_for_each_surface (PhocOutput *self, PhocView *view,
-                                   PhocSurfaceIterator iterator, void
-                                   *user_data)
+phoc_output_view_for_each_surface (PhocOutput          *self,
+                                   PhocView            *view,
+                                   PhocSurfaceIterator  iterator,
+                                   void                *user_data)
 {
   struct wlr_box *output_box =
     wlr_output_layout_get_box (self->desktop->layout, self->wlr_output);
@@ -515,11 +680,20 @@ phoc_output_view_for_each_surface (PhocOutput *self, PhocView *view,
 }
 
 #ifdef PHOC_XWAYLAND
+/**
+ * phoc_output_xwayland_children_for_each_surface:
+ * @self: the output
+ * @surface: The wlr_xwayland_surface
+ * @iterator: (scope call): The callback invoked on each iteration
+ * @user_data: Callback user data
+ *
+ * Iterate over children of a `wlr_xwayland_surface`.
+ */
 void
-phoc_output_xwayland_children_for_each_surface (PhocOutput *self, struct
-                                                wlr_xwayland_surface *surface,
-                                                PhocSurfaceIterator
-                                                iterator, void *user_data)
+phoc_output_xwayland_children_for_each_surface (PhocOutput                  *self,
+                                                struct wlr_xwayland_surface *surface,
+                                                PhocSurfaceIterator          iterator,
+                                                void                        *user_data)
 {
   struct wlr_box *output_box =
     wlr_output_layout_get_box (self->desktop->layout, self->wlr_output);
@@ -566,40 +740,65 @@ phoc_output_layer_handle_surface (PhocOutput *self, PhocLayerSurface *layer_surf
 
     double popup_sx, popup_sy;
     popup_sx = layer_surface->geo.x;
-    popup_sx += popup->popup->geometry.x - popup->geometry.x;
+    popup_sx += popup->popup->geometry.x - popup->current.geometry.x;
     popup_sy = layer_surface->geo.y;
-    popup_sy += popup->popup->geometry.y - popup->geometry.y;
+    popup_sy += popup->popup->geometry.y - popup->current.geometry.y;
 
     phoc_output_xdg_surface_for_each_surface (self, popup,
                                               popup_sx, popup_sy, iterator, user_data);
   }
 }
 
+/**
+ * phoc_output_layer_for_each_surface:
+ * @self: the output
+ * @layer: The layer that should be iterated over
+ * @iterator: (scope call): The callback invoked on each iteration
+ * @user_data: Callback user data
+ *
+ * Iterate over [type@LayerSurface]s in a layer.
+ */
 void
-phoc_output_layer_for_each_surface (PhocOutput                   *self,
-                                    struct wl_list               *layer_surfaces,
-                                    PhocSurfaceIterator iterator,
-                                    void                         *user_data)
+phoc_output_layer_for_each_surface (PhocOutput          *self,
+                                    enum zwlr_layer_shell_v1_layer layer,
+                                    PhocSurfaceIterator  iterator,
+                                    void                *user_data)
 {
   PhocLayerSurface *layer_surface;
 
-  wl_list_for_each_reverse (layer_surface, layer_surfaces, link)
+  wl_list_for_each_reverse (layer_surface, &self->layer_surfaces, link)
   {
+    if (layer_surface->layer != layer)
+      continue;
+
     if (layer_surface->layer_surface->current.exclusive_zone <= 0) {
       phoc_output_layer_handle_surface (self, layer_surface, iterator, user_data);
     }
   }
-  wl_list_for_each (layer_surface, layer_surfaces, link) {
+  wl_list_for_each (layer_surface, &self->layer_surfaces, link) {
+    if (layer_surface->layer != layer)
+      continue;
+
     if (layer_surface->layer_surface->current.exclusive_zone > 0) {
       phoc_output_layer_handle_surface (self, layer_surface, iterator, user_data);
     }
   }
 }
 
+/**
+ * phoc_output_drag_icons_for_each_surface:
+ * @self: the output
+ * @input: a [type@Input]
+ * @iterator: (scope call): The iterator
+ * @user_data: Callback user data
+ *
+ * Iterate over the surface tree of the drag icon's surface of an input's seats.
+ */
 void
-phoc_output_drag_icons_for_each_surface (PhocOutput *self, PhocInput *input,
-                                         PhocSurfaceIterator
-                                         iterator, void *user_data)
+phoc_output_drag_icons_for_each_surface (PhocOutput          *self,
+                                         PhocInput           *input,
+                                         PhocSurfaceIterator  iterator,
+                                         void                *user_data)
 {
   struct wlr_box *output_box =
     wlr_output_layout_get_box (self->desktop->layout, self->wlr_output);
@@ -624,9 +823,20 @@ phoc_output_drag_icons_for_each_surface (PhocOutput *self, PhocInput *input,
   }
 }
 
+/**
+ * phoc_output_for_each_surface:
+ * @self: the output
+ * @iterator: (scope call): The iterator
+ * @user_data: Callback user data
+ * @visible_only: Whether to only iterate over visible surfaces
+ *
+ * Iterate over surfaces on the output.
+ */
 void
-phoc_output_for_each_surface (PhocOutput *self, PhocSurfaceIterator iterator, void
-                              *user_data, gboolean visible_only)
+phoc_output_for_each_surface (PhocOutput          *self,
+                              PhocSurfaceIterator  iterator,
+                              void                *user_data,
+                              gboolean             visible_only)
 {
   PhocDesktop *desktop = self->desktop;
   PhocServer *server = phoc_server_get_default ();
@@ -658,9 +868,9 @@ phoc_output_for_each_surface (PhocOutput *self, PhocSurfaceIterator iterator, vo
   phoc_output_drag_icons_for_each_surface (self, server->input,
                                            iterator, user_data);
 
-  for (size_t i = 0; i < G_N_ELEMENTS (self->layers); ++i) {
-    phoc_output_layer_for_each_surface (self, &self->layers[i],
-                                        iterator, user_data);
+  for (enum zwlr_layer_shell_v1_layer layer = ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND;
+       layer <= ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY; layer++) {
+    phoc_output_layer_for_each_surface (self, layer, iterator, user_data);
   }
 }
 
@@ -680,8 +890,7 @@ phoc_output_scale_box (PhocOutput *self, struct wlr_box *box, float scale)
 }
 
 void
-phoc_output_get_decoration_box (PhocOutput *self, PhocView *view,
-                                struct wlr_box *box)
+phoc_output_get_decoration_box (PhocOutput *self, PhocView *view, struct wlr_box *box)
 {
   struct wlr_box deco_box;
 
@@ -752,27 +961,25 @@ damage_surface_iterator (PhocOutput *self, struct wlr_surface *surface, struct
   int center_x = box.x + box.width/2;
   int center_y = box.y + box.height/2;
 
-  if (pixman_region32_not_empty (&surface->buffer_damage)) {
-    pixman_region32_t damage;
-    pixman_region32_init (&damage);
-    wlr_surface_get_effective_damage (surface, &damage);
-    wlr_region_scale (&damage, &damage, scale);
-    wlr_region_scale (&damage, &damage, self->wlr_output->scale);
-    if (ceil (self->wlr_output->scale) > surface->current.scale) {
-      // When scaling up a surface, it'll become blurry so we need to
-      // expand the damage region
-      wlr_region_expand (&damage, &damage,
-                         ceil (self->wlr_output->scale) - surface->current.scale);
-    }
-    pixman_region32_translate (&damage, box.x, box.y);
-    wlr_region_rotated_bounds (&damage, &damage, rotation,
-                               center_x, center_y);
-    wlr_output_damage_add (self->damage, &damage);
-    pixman_region32_fini (&damage);
+  pixman_region32_t damage;
+  pixman_region32_init (&damage);
+  wlr_surface_get_effective_damage (surface, &damage);
+  wlr_region_scale (&damage, &damage, scale);
+  wlr_region_scale (&damage, &damage, self->wlr_output->scale);
+  if (ceil (self->wlr_output->scale) > surface->current.scale) {
+    // When scaling up a surface, it'll become blurry so we need to
+    // expand the damage region
+    wlr_region_expand (&damage, &damage,
+                       ceil (self->wlr_output->scale) - surface->current.scale);
   }
+  pixman_region32_translate (&damage, box.x, box.y);
+  wlr_region_rotated_bounds (&damage, &damage, rotation,
+                             center_x, center_y);
+  wlr_output_damage_add (self->damage, &damage);
+  pixman_region32_fini (&damage);
 
   if (*whole) {
-    wlr_box_rotated_bounds (&box, &box, rotation);
+    phoc_utils_rotated_bounds (&box, &box, rotation);
     wlr_output_damage_add_box (self->damage, &box);
   }
 
@@ -1011,7 +1218,7 @@ phoc_output_is_match (PhocOutput *self,
   return match;
 }
 
-/*
+/**
  * phoc_output_has_fullscreen_view:
  * @self: The #PhocOutput
  *
@@ -1025,3 +1232,215 @@ phoc_output_has_fullscreen_view (PhocOutput *self)
 
   return self->fullscreen_view != NULL && self->fullscreen_view->wlr_surface != NULL;
 }
+
+
+guint
+phoc_output_add_frame_callback  (PhocOutput        *self,
+                                 PhocAnimatable    *animatable,
+                                 PhocFrameCallback  callback,
+                                 gpointer           user_data,
+                                 GDestroyNotify     notify)
+{
+  PhocOutputPrivate *priv;
+  PhocOutputFrameCallbackInfo *cb_info = g_new0 (PhocOutputFrameCallbackInfo, 1);
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  *cb_info = (PhocOutputFrameCallbackInfo) {
+    .animatable = animatable,
+    .callback = callback,
+    .user_data = user_data,
+    .notify = notify,
+    .id = priv->frame_callback_next_id,
+  };
+
+  if (priv->frame_callbacks == NULL) {
+    priv->last_frame_us = g_get_monotonic_time ();
+    /* No other frame callbacks so need to schedule a frame to keep
+     * frame clock ticking */
+    wlr_output_schedule_frame (self->wlr_output);
+  }
+
+  priv->frame_callbacks = g_slist_prepend (priv->frame_callbacks, cb_info);
+  return priv->frame_callback_next_id++;
+}
+
+
+void
+phoc_output_remove_frame_callback  (PhocOutput *self, guint id)
+{
+  PhocOutputPrivate *priv;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  for (GSList *elem = priv->frame_callbacks; elem; elem = elem->next) {
+    PhocOutputFrameCallbackInfo *cb_info = elem->data;
+
+    if (cb_info->id == id) {
+      phoc_output_frame_callback_info_free (cb_info);
+      priv->frame_callbacks = g_slist_delete_link (priv->frame_callbacks, elem);
+      return;
+    }
+  }
+  g_return_if_reached();
+}
+
+
+bool
+phoc_output_has_frame_callbacks  (PhocOutput *self)
+{
+  PhocOutputPrivate *priv;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  return !!priv->frame_callbacks;
+}
+
+
+void
+phoc_output_lower_shield (PhocOutput *self)
+{
+  PhocOutputPrivate *priv;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  if (priv->shield == NULL)
+    return;
+
+  phoc_output_shield_lower (priv->shield);
+}
+
+
+void
+phoc_output_raise_shield (PhocOutput *self)
+{
+  PhocOutputPrivate *priv;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  if (priv->shield == NULL)
+    priv->shield = phoc_output_shield_new (self);
+
+  phoc_output_shield_raise (priv->shield);
+}
+
+/**
+ * phoc_output_has_layer:
+ * @self: The #PhocOutput
+ *
+ * Returns: %TRUE if the output has a Phoc.LayerSurface in the given layer.
+ *          %FALSE otherwise.
+ */
+gboolean
+phoc_output_has_layer (PhocOutput *self, enum zwlr_layer_shell_v1_layer layer)
+{
+  PhocLayerSurface *layer_surface;
+
+  g_assert (PHOC_IS_OUTPUT (self));
+
+  wl_list_for_each (layer_surface, &self->layer_surfaces, link) {
+    if (layer_surface->layer == layer)
+      return TRUE;
+  }
+
+  return FALSE;
+}
+
+
+static gboolean
+should_reveal_shell (PhocOutput *self)
+{
+  PhocOutputPrivate *priv;
+  PhocServer *server = phoc_server_get_default();
+  PhocLayerSurface *layer_surface;
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  /* is our layer-surface focused on some seat? */
+  for (GSList *elem = phoc_input_get_seats (server->input); elem; elem = elem->next) {
+    PhocSeat *seat = PHOC_SEAT (elem->data);
+    if (seat->focused_layer && seat->focused_layer->output == self->wlr_output) {
+      return true;
+    }
+  }
+
+  /* is some draggable surface unfolded, being dragged or animated? */
+  wl_list_for_each (layer_surface, &self->layer_surfaces, link) {
+    PhocDraggableLayerSurface *draggable =
+        phoc_desktop_get_draggable_layer_surface (server->desktop, layer_surface);
+    if (draggable &&
+        (phoc_draggable_layer_surface_get_state (draggable) != PHOC_DRAGGABLE_SURFACE_STATE_NONE ||
+         phoc_draggable_layer_surface_is_unfolded (draggable))) {
+      return true;
+    }
+  }
+
+  /* is shell reveal forced by user gesture? */
+  return priv->force_shell_reveal;
+}
+
+
+/**
+ * phoc_output_update_shell_reveal:
+ * @self: The #PhocOutput
+ *
+ * Updates shell reveal status of given output.
+ */
+void
+phoc_output_update_shell_reveal (PhocOutput *self)
+{
+  PhocOutputPrivate *priv;
+  gboolean old;
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  old = priv->shell_revealed;
+  priv->shell_revealed = should_reveal_shell (self);
+
+  if (priv->shell_revealed != old) {
+    phoc_output_damage_whole (self);
+  }
+}
+
+
+/**
+ * phoc_output_force_shell_reveal:
+ * @self: The #PhocOutput
+ * @force: %TRUE to forcefully reveal shell; %FALSE otherwise.
+ *
+ * Allows to force shell reveal status regardless of whether
+ * other conditions are being fulfilled.
+ */
+void
+phoc_output_force_shell_reveal (PhocOutput *self, gboolean force)
+{
+  PhocOutputPrivate *priv;
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+
+  priv->force_shell_reveal = force;
+  phoc_output_update_shell_reveal (self);
+}
+
+
+/**
+ * phoc_output_has_shell_revealed:
+ * @self: The #PhocOutput
+ *
+ * Returns: %TRUE if layer-shell's TOP layer should appear on top
+ *          of fullscreen windows on this output.
+ *          %FALSE otherwise.
+ */
+gboolean
+phoc_output_has_shell_revealed (PhocOutput *self)
+{
+  PhocOutputPrivate *priv;
+  g_assert (PHOC_IS_OUTPUT (self));
+  priv = phoc_output_get_instance_private (self);
+  return priv->shell_revealed;
+}
