@@ -15,7 +15,9 @@
 #include "seat.h"
 #include "server.h"
 #include "render.h"
+#include "render-private.h"
 #include "xwayland-surface.h"
+#include "utils.h"
 
 #define _POSIX_C_SOURCE 200809L
 #include <assert.h>
@@ -26,6 +28,7 @@
 #include <time.h>
 #include <wlr/backend.h>
 #include <wlr/config.h>
+#include <wlr/render/drm_format_set.h>
 #include <wlr/render/wlr_renderer.h>
 #include <wlr/render/gles2.h>
 #include <wlr/render/egl.h>
@@ -35,15 +38,19 @@
 #include <wlr/types/wlr_linux_dmabuf_v1.h>
 #include <wlr/util/region.h>
 #include <wlr/version.h>
+#include <wlr/render/allocator.h>
 #include <GLES2/gl2.h>
 #include <GLES2/gl2ext.h>
 
-/* Private wlr_allocator headers */
-struct wlr_gbm_allocator *wlr_gbm_allocator_create(int drm_fd);
-struct wlr_buffer *wlr_allocator_create_buffer (struct wlr_allocator *alloc, int width, int height, const struct wlr_drm_format *format);
-void wlr_allocator_destroy (struct wlr_allocator *alloc);
-struct wlr_drm_format *wlr_drm_format_create (uint32_t format);
-/* ----------------------------- */
+
+static void wlr_box_from_pixman_box32(struct wlr_box *dest, const pixman_box32_t box) {
+	*dest = (struct wlr_box){
+		.x = box.x1,
+		.y = box.y1,
+		.width = box.x2 - box.x1,
+		.height = box.y2 - box.y1,
+	};
+}
 
 #define TOUCH_POINT_SIZE 20
 #define TOUCH_POINT_BORDER 0.1
@@ -70,7 +77,7 @@ static guint signals[N_SIGNALS] = { 0 };
 
 enum {
   PROP_0,
-  PROP_WLR_RENDERER,
+  PROP_WLR_BACKEND,
   PROP_LAST_PROP
 };
 static GParamSpec *props[PROP_LAST_PROP];
@@ -78,10 +85,15 @@ static GParamSpec *props[PROP_LAST_PROP];
 struct _PhocRenderer {
   GObject               parent;
 
+  struct wlr_backend   *wlr_backend;
   struct wlr_renderer  *wlr_renderer;
-  struct wlr_allocator *allocator;
+  struct wlr_allocator *wlr_allocator;
 };
-G_DEFINE_TYPE (PhocRenderer, phoc_renderer, G_TYPE_OBJECT)
+
+static void phoc_renderer_initable_iface_init (GInitableIface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (PhocRenderer, phoc_renderer, G_TYPE_OBJECT,
+                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, phoc_renderer_initable_iface_init));
 
 
 struct render_data {
@@ -111,8 +123,8 @@ phoc_renderer_set_property (GObject      *object,
   PhocRenderer *self = PHOC_RENDERER (object);
 
   switch (property_id) {
-  case PROP_WLR_RENDERER:
-    self->wlr_renderer = g_value_get_pointer (value);
+  case PROP_WLR_BACKEND:
+    self->wlr_backend = g_value_get_pointer (value);
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -130,8 +142,8 @@ phoc_renderer_get_property (GObject    *object,
   PhocRenderer *self = PHOC_RENDERER (object);
 
   switch (property_id) {
-  case PROP_WLR_RENDERER:
-    g_value_set_pointer (value, self->wlr_renderer);
+  case PROP_WLR_BACKEND:
+    g_value_set_pointer (value, self->wlr_backend);
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
@@ -142,10 +154,6 @@ phoc_renderer_get_property (GObject    *object,
 
 static void scissor_output(struct wlr_output *wlr_output,
 		pixman_box32_t *rect) {
-	struct wlr_renderer *renderer =
-		wlr_backend_get_renderer(wlr_output->backend);
-	assert(renderer);
-
 	struct wlr_box box = {
 		.x = rect->x1,
 		.y = rect->y1,
@@ -160,7 +168,7 @@ static void scissor_output(struct wlr_output *wlr_output,
 		wlr_output_transform_invert(wlr_output->transform);
 	wlr_box_transform(&box, &box, transform, ow, oh);
 
-	wlr_renderer_scissor(renderer, &box);
+	wlr_renderer_scissor(wlr_output->renderer, &box);
 }
 
 static void render_texture(struct wlr_output *wlr_output,
@@ -168,17 +176,13 @@ static void render_texture(struct wlr_output *wlr_output,
 		const struct wlr_fbox *src_box, const struct wlr_box *dst_box,
 		const float matrix[static 9],
 		float rotation, float alpha) {
-	struct wlr_renderer *renderer =
-		wlr_backend_get_renderer(wlr_output->backend);
-	assert(renderer);
-
 	struct wlr_box rotated;
-	wlr_box_rotated_bounds(&rotated, dst_box, rotation);
+	phoc_utils_rotated_bounds(&rotated, dst_box, rotation);
 
 	pixman_region32_t damage;
 	pixman_region32_init(&damage);
-	pixman_region32_union_rect(&damage, &damage, rotated.x, rotated.y,
-		rotated.width, rotated.height);
+	pixman_region32_union_rect(&damage, &damage, dst_box->x, dst_box->y,
+		dst_box->width, dst_box->height);
 	pixman_region32_intersect(&damage, &damage, output_damage);
 	bool damaged = pixman_region32_not_empty(&damage);
 	if (!damaged) {
@@ -191,9 +195,11 @@ static void render_texture(struct wlr_output *wlr_output,
 		scissor_output(wlr_output, &rects[i]);
 
 		if (src_box != NULL) {
-			wlr_render_subtexture_with_matrix(renderer, texture, src_box, matrix, alpha);
+			wlr_render_subtexture_with_matrix(wlr_output->renderer,
+                                                          texture, src_box, matrix, alpha);
 		} else {
-			wlr_render_texture_with_matrix(renderer, texture, matrix, alpha);
+			wlr_render_texture_with_matrix(wlr_output->renderer,
+                                                       texture, matrix, alpha);
 		}
 	}
 
@@ -226,7 +232,7 @@ collect_touch_points (PhocOutput *output, struct wlr_surface *surface, struct wl
 }
 
 static void render_surface_iterator(PhocOutput *output,
-		struct wlr_surface *surface, struct wlr_box *_box, float rotation,
+		struct wlr_surface *surface, struct wlr_box *box, float rotation,
 		float scale, void *_data) {
 	struct render_data *data = _data;
 	struct wlr_output *wlr_output = output->wlr_output;
@@ -241,9 +247,9 @@ static void render_surface_iterator(PhocOutput *output,
 	struct wlr_fbox src_box;
 	wlr_surface_get_buffer_source_box(surface, &src_box);
 
-	struct wlr_box dst_box = *_box;
-	phoc_output_scale_box(wlr_output->data, &dst_box, scale);
-	phoc_output_scale_box(wlr_output->data, &dst_box, wlr_output->scale);
+	struct wlr_box dst_box = *box;
+	phoc_output_scale_box (output, &dst_box, scale);
+	phoc_output_scale_box (output, &dst_box, wlr_output->scale);
 
 	float matrix[9];
 	enum wl_output_transform transform =
@@ -266,10 +272,6 @@ static void render_decorations(PhocOutput *output,
 		return;
 	}
 
-	struct wlr_renderer *renderer =
-		wlr_backend_get_renderer(output->wlr_output->backend);
-	assert(renderer);
-
 	struct wlr_box box;
 	phoc_output_get_decoration_box(output, view, &box);
 
@@ -293,7 +295,7 @@ static void render_decorations(PhocOutput *output,
 		pixman_region32_rectangles(&damage, &nrects);
 	for (int i = 0; i < nrects; ++i) {
 		scissor_output(output->wlr_output, &rects[i]);
-		wlr_render_quad_with_matrix(renderer, color, matrix);
+		wlr_render_quad_with_matrix(output->wlr_output->renderer, color, matrix);
 	}
 
 buffer_damage_finish:
@@ -314,18 +316,21 @@ static void render_view(PhocOutput *output, PhocView *view,
 	phoc_output_view_for_each_surface(output, view, render_surface_iterator, data);
 }
 
-static void render_layer(PhocOutput *output,
-		pixman_region32_t *damage, struct wl_list *layer_surfaces) {
-	struct render_data data = {
-		.damage = damage,
-		.alpha = 1.0f,
-	};
-	phoc_output_layer_for_each_surface(output, layer_surfaces,
-		render_surface_iterator, &data);
+static void
+render_layer (PhocOutput                     *output,
+              pixman_region32_t              *damage,
+              enum zwlr_layer_shell_v1_layer  layer)
+{
+  struct render_data data = {
+    .damage = damage,
+    .alpha = 1.0f,
+  };
+
+  phoc_output_layer_for_each_surface(output, layer, render_surface_iterator, &data);
 }
 
 static void count_surface_iterator(PhocOutput *output,
-		struct wlr_surface *surface, struct wlr_box *_box, float rotation,
+		struct wlr_surface *surface, struct wlr_box *box, float rotation,
 		float scale, void *data) {
 	size_t *n = data;
 	n++;
@@ -345,9 +350,8 @@ static bool scan_out_fullscreen_view(PhocOutput *output) {
 		}
 	}
 
-	if (!wl_list_empty(&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY])) {
+	if (phoc_output_has_layer (output, ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY))
 		return false;
-	}
 
 	struct wlr_output_cursor *cursor;
 	wl_list_for_each(cursor, &wlr_output->cursors, link) {
@@ -453,24 +457,24 @@ render_touch_point_cb (gpointer data, gpointer user_data)
 
   PhocOutput *output = user_data;
   struct wlr_output *wlr_output = output->wlr_output;
+  struct wlr_renderer *wlr_renderer = wlr_output->renderer;
 
   int size = TOUCH_POINT_SIZE * wlr_output->scale;
   struct wlr_box point_box = wlr_box_from_touch_point (touch_point, size, size);
 
-  struct wlr_renderer *renderer = wlr_backend_get_renderer (wlr_output->backend);
-
   float color[4] = {touch_point->id * 100 + 240, 1.0, 1.0, 0.75};
   color_hsv_to_rgb (color);
-  wlr_render_rect (renderer, &point_box, color, wlr_output->transform_matrix);
+  wlr_render_rect (wlr_renderer, &point_box, color, wlr_output->transform_matrix);
 
   size = TOUCH_POINT_SIZE * (1.0 - TOUCH_POINT_BORDER) * wlr_output->scale;
   point_box = wlr_box_from_touch_point (touch_point, size, size);
-  wlr_render_rect (renderer, &point_box, (float[])COLOR_TRANSPARENT_WHITE, wlr_output->transform_matrix);
+  wlr_render_rect (wlr_renderer, &point_box,
+                   (float[])COLOR_TRANSPARENT_WHITE, wlr_output->transform_matrix);
 
   point_box = wlr_box_from_touch_point (touch_point, 8 * wlr_output->scale, 2 * wlr_output->scale);
-  wlr_render_rect (renderer, &point_box, color, wlr_output->transform_matrix);
+  wlr_render_rect (wlr_renderer, &point_box, color, wlr_output->transform_matrix);
   point_box = wlr_box_from_touch_point (touch_point, 2 * wlr_output->scale, 8 * wlr_output->scale);
-  wlr_render_rect (renderer, &point_box, color, wlr_output->transform_matrix);
+  wlr_render_rect (wlr_renderer, &point_box, color, wlr_output->transform_matrix);
 }
 
 static void
@@ -518,31 +522,35 @@ view_render_iterator (struct wlr_surface *surface, int sx, int sy, void *_data)
 
   PhocServer *server = phoc_server_get_default ();
   PhocRenderer *self = phoc_server_get_renderer (server);
-  struct wlr_texture *view_texture = wlr_surface_get_texture (surface);
+  struct wlr_texture *texture = wlr_surface_get_texture (surface);
 
   struct view_render_data *data = _data;
   PhocView *view = data->view;
-  struct wlr_surface *root = view->wlr_surface;
-
-  struct wlr_box box;
-  view_get_box (view, &box);
 
   struct wlr_box geo;
   view_get_geometry (view, &geo);
 
-  float mat[16];
-  wlr_matrix_identity (mat);
+  float scale = fmin (data->width / (float)geo.width,
+                      data->height / (float)geo.height);
 
-  wlr_matrix_scale (mat, data->width / (float)geo.width, data->height / (float)geo.height);
-  wlr_matrix_scale (mat, 1 / (float)root->current.scale, 1 / (float)root->current.scale);
-  wlr_matrix_scale (mat, view->scale, view->scale);
-  wlr_matrix_scale (mat, root->current.scale / surface->current.scale, root->current.scale / surface->current.scale);
-  wlr_matrix_translate (mat, -geo.x, -geo.y);
+  float proj[9];
+  wlr_matrix_identity (proj);
+  wlr_matrix_scale (proj, scale, scale);
+  wlr_matrix_translate (proj, -geo.x, -geo.y);
 
-  wlr_render_texture (self->wlr_renderer,
-                      view_texture, mat, sx * surface->current.scale,
-                      sy * surface->current.scale,
-                      1.0);
+  struct wlr_fbox src_box;
+  wlr_surface_get_buffer_source_box (surface, &src_box);
+
+  struct wlr_box dst_box = {
+    .x = sx,
+    .y = sy,
+    .width = surface->current.width,
+    .height = surface->current.height,
+  };
+
+  float mat[9];
+  wlr_matrix_project_box (mat, &dst_box, wlr_output_transform_invert (surface->current.transform), 0, proj);
+  wlr_render_subtexture_with_matrix (self->wlr_renderer, texture, &src_box, mat, 1.0);
 }
 
 gboolean
@@ -555,15 +563,19 @@ phoc_renderer_render_view_to_buffer (PhocRenderer *self,
                                      void         *data)
 {
   struct wlr_surface *surface = view->wlr_surface;
+  struct wlr_buffer *buffer;
 
   g_return_val_if_fail (surface, false);
-  g_return_val_if_fail (self->allocator, false);
+  g_return_val_if_fail (self->wlr_allocator, false);
 
-  struct wlr_drm_format *fmt = wlr_drm_format_create (DRM_FORMAT_ARGB8888);
-  struct wlr_buffer *buffer = wlr_allocator_create_buffer (self->allocator, width, height, fmt);
+  struct wlr_drm_format_set fmt_set = {};
+  wlr_drm_format_set_add (&fmt_set, DRM_FORMAT_ARGB8888, DRM_FORMAT_MOD_INVALID);
 
+  const struct wlr_drm_format *fmt = wlr_drm_format_set_get (&fmt_set, DRM_FORMAT_ARGB8888);
+
+  buffer = wlr_allocator_create_buffer (self->wlr_allocator, width, height, fmt);
   if (!buffer) {
-    free (fmt);
+    wlr_drm_format_set_finish (&fmt_set);
     g_return_val_if_reached (false);
   }
 
@@ -580,7 +592,7 @@ phoc_renderer_render_view_to_buffer (PhocRenderer *self,
   wlr_renderer_end (self->wlr_renderer);
 
   wlr_buffer_drop (buffer);
-  free(fmt);
+  wlr_drm_format_set_finish (&fmt_set);
 
   return true;
 }
@@ -653,16 +665,8 @@ void phoc_renderer_render_output (PhocRenderer *self, PhocOutput *output) {
 
 	// Check if we can delegate the fullscreen surface to the output
 	if (phoc_output_has_fullscreen_view (output)) {
-		static bool last_scanned_out = false;
 		bool scanned_out = scan_out_fullscreen_view(output);
 
-		if (scanned_out && !last_scanned_out) {
-			g_debug ("Starting fullscreen view scan out");
-		} else if (!scanned_out && last_scanned_out) {
-			g_debug ("Stopping fullscreen view scan out");
-		}
-		last_scanned_out = scanned_out;
-
 		if (scanned_out) {
 			goto send_frame_done;
 		}
@@ -732,17 +736,14 @@ void phoc_renderer_render_output (PhocRenderer *self, PhocOutput *output) {
 		}
 #endif
 
-		if (output->force_shell_reveal) {
+		if (phoc_output_has_shell_revealed (output)) {
 			// Render top layer above fullscreen view when requested
-			render_layer(output, &buffer_damage,
-				&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_TOP]);
+			render_layer (output, &buffer_damage, ZWLR_LAYER_SHELL_V1_LAYER_TOP);
 		}
 	} else {
 		// Render background and bottom layers under views
-		render_layer(output, &buffer_damage,
-			&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND]);
-		render_layer(output, &buffer_damage,
-			&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM]);
+		render_layer (output, &buffer_damage, ZWLR_LAYER_SHELL_V1_LAYER_BACKGROUND);
+		render_layer (output, &buffer_damage, ZWLR_LAYER_SHELL_V1_LAYER_BOTTOM);
 
 		PhocView *view;
 			// Render all views
@@ -753,14 +754,12 @@ void phoc_renderer_render_output (PhocRenderer *self, PhocOutput *output) {
 		}
 
 		// Render top layer above views
-		render_layer(output, &buffer_damage,
-			&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_TOP]);
+		render_layer (output, &buffer_damage, ZWLR_LAYER_SHELL_V1_LAYER_TOP);
 	}
 
 	render_drag_icons(output, &buffer_damage, server->input);
 
-	render_layer(output, &buffer_damage,
-		&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY]);
+	render_layer (output, &buffer_damage, ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY);
 
 renderer_end:
 	wlr_output_render_software_cursors(wlr_output, &buffer_damage);
@@ -801,14 +800,31 @@ send_frame_done:
 }
 
 
-static void
-phoc_renderer_constructed (GObject *object)
+static gboolean
+phoc_renderer_initable_init (GInitable    *initable,
+                             GCancellable *cancellable,
+                             GError      **error)
 {
-  PhocRenderer *self = PHOC_RENDERER (object);
-  int drm_fd = fcntl (wlr_renderer_get_drm_fd (self->wlr_renderer), F_DUPFD_CLOEXEC, 0);
-  // TODO: use wlr_allocator_autocreate or retrieve a reference from wlr_renderer once possible
-  if (wlr_renderer_is_gles2 (self->wlr_renderer))
-    self->allocator = (struct wlr_allocator*) wlr_gbm_allocator_create (drm_fd);
+  PhocRenderer *self = PHOC_RENDERER (initable);
+
+  self->wlr_renderer = wlr_renderer_autocreate (self->wlr_backend);
+  if (self->wlr_renderer == NULL) {
+    g_set_error (error,
+                 G_FILE_ERROR, G_FILE_ERROR_FAILED,
+		 "Could not create renderer");
+    return FALSE;
+  }
+
+  self->wlr_allocator = wlr_allocator_autocreate (self->wlr_backend,
+                                                  self->wlr_renderer);
+  if (self->wlr_allocator == NULL) {
+    g_set_error (error,
+                 G_FILE_ERROR, G_FILE_ERROR_FAILED,
+		 "Could not create allocator");
+    return FALSE;
+  }
+
+  return TRUE;
 }
 
 
@@ -816,12 +832,19 @@ static void
 phoc_renderer_finalize (GObject *object)
 {
   PhocRenderer *self = PHOC_RENDERER (object);
-  if (self->allocator)
-    wlr_allocator_destroy (self->allocator);
+  if (self->wlr_allocator)
+    wlr_allocator_destroy (self->wlr_allocator);
   /* TODO: destroy wlr_renderer */
 }
 
 
+static void
+phoc_renderer_initable_iface_init (GInitableIface *iface)
+{
+  iface->init = phoc_renderer_initable_init;
+}
+
+
 static void
 phoc_renderer_class_init (PhocRendererClass *klass)
 {
@@ -829,16 +852,15 @@ phoc_renderer_class_init (PhocRendererClass *klass)
 
   object_class->get_property = phoc_renderer_get_property;
   object_class->set_property = phoc_renderer_set_property;
-  object_class->constructed = phoc_renderer_constructed;
   object_class->finalize = phoc_renderer_finalize;
 
   /**
-   * PhocRenderer:wlr-renderer
+   * PhocRenderer:wlr-backend
    *
-   * The wlr-renderer from wlroots to use
+   * The wlr-backend to use for initializing the renderer
    */
-  props[PROP_WLR_RENDERER] =
-    g_param_spec_pointer ("wlr-renderer",
+  props[PROP_WLR_BACKEND] =
+    g_param_spec_pointer ("wlr-backend",
                           "",
                           "",
                           G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
@@ -879,9 +901,27 @@ phoc_renderer_init (PhocRenderer *self)
 
 
 PhocRenderer *
-phoc_renderer_new (struct wlr_renderer *wlr_renderer)
+phoc_renderer_new (struct wlr_backend *wlr_backend, GError **error)
 {
-  return PHOC_RENDERER (g_object_new (PHOC_TYPE_RENDERER,
-                                      "wlr-renderer", wlr_renderer,
-                                      NULL));
+  return PHOC_RENDERER (g_initable_new (PHOC_TYPE_RENDERER, NULL, error,
+                                        "wlr-backend", wlr_backend,
+                                        NULL));
+}
+
+
+struct wlr_renderer *
+phoc_renderer_get_wlr_renderer (PhocRenderer *self)
+{
+  g_assert (PHOC_IS_RENDERER (self));
+
+  return self->wlr_renderer;
+}
+
+
+struct wlr_allocator *
+phoc_renderer_get_wlr_allocator (PhocRenderer *self)
+{
+  g_assert (PHOC_IS_RENDERER (self));
+
+  return self->wlr_allocator;
 }
