@@ -5,8 +5,6 @@
  */
 #pragma once
 
-#include "output.h"
-
 #include <glib-object.h>
 
 #include <wlr/render/wlr_renderer.h>
@@ -17,7 +15,11 @@ G_BEGIN_DECLS
 
 G_DECLARE_FINAL_TYPE (PhocRenderer, phoc_renderer, PHOC, RENDERER, GObject)
 
-PhocRenderer *phoc_renderer_new (struct wlr_renderer *wlr_renderer);
+typedef struct _PhocOutput PhocOutput;
+typedef struct _PhocView PhocView;
+
+PhocRenderer *phoc_renderer_new (struct wlr_backend *wlr_backend, GError **error);
+
 void          phoc_renderer_render_output (PhocRenderer *self, PhocOutput *output);
 gboolean      phoc_renderer_render_view_to_buffer (PhocRenderer *self,
                                                    PhocView     *view,
@@ -26,5 +28,4 @@ gboolean      phoc_renderer_render_view_to_buffer (PhocRenderer *self,
                                                    int           stride,
                                                    uint32_t     *flags,
                                                    void         *data);
-
 G_END_DECLS
