@@ -170,14 +170,14 @@ static void
 handle_switch_toggle (struct wl_listener *listener, void *data)
 {
   PhocServer *server = phoc_server_get_default ();
-  struct roots_switch *switch_device =
+  struct phoc_switch *switch_device =
     wl_container_of (listener, switch_device, toggle);
   PhocDesktop *desktop = server->desktop;
 
   wlr_idle_notify_activity (desktop->idle, switch_device->seat->seat);
   struct wlr_event_switch_toggle *event = data;
 
-  roots_switch_handle_toggle (switch_device, event);
+  phoc_switch_handle_toggle (switch_device, event);
 }
 
 static void
@@ -206,16 +206,11 @@ handle_touch_up (struct wl_listener *listener, void *data)
   PhocCursor *cursor = wl_container_of (listener, cursor, touch_up);
   struct wlr_event_touch_up *event = data;
   PhocDesktop *desktop = server->desktop;
-  PhocOutput *output = g_hash_table_lookup (desktop->input_output_map,
-                                            event->device->name);
 
-  /* handle touch up regardless of output status so events don't become stuck */
-  phoc_cursor_handle_touch_up (cursor, event);
-  if (output && !output->wlr_output->enabled) {
-    g_debug ("Touch event ignored since output '%s' is disabled.",
-             output->wlr_output->name);
+  if (!phoc_cursor_is_active_touch_id (cursor, event->touch_id)) {
     return;
   }
+  phoc_cursor_handle_touch_up (cursor, event);
   wlr_idle_notify_activity (desktop->idle, cursor->seat->seat);
 }
 
@@ -226,17 +221,11 @@ handle_touch_motion (struct wl_listener *listener, void *data)
   PhocCursor *cursor = wl_container_of (listener, cursor, touch_motion);
   struct wlr_event_touch_motion *event = data;
   PhocDesktop *desktop = server->desktop;
-  PhocOutput *output = g_hash_table_lookup (desktop->input_output_map,
-                                            event->device->name);
 
-  /* handle touch motion regardless of output status so events don't become
-     stuck */
-  phoc_cursor_handle_touch_motion (cursor, event);
-  if (output && !output->wlr_output->enabled) {
-    g_debug ("Touch event ignored since output '%s' is disabled.",
-             output->wlr_output->name);
+  if (!phoc_cursor_is_active_touch_id (cursor, event->touch_id)) {
     return;
   }
+  phoc_cursor_handle_touch_motion (cursor, event);
   wlr_idle_notify_activity (desktop->idle, cursor->seat->seat);
 }
 
@@ -800,8 +789,6 @@ phoc_drag_icon_update_position (PhocDragIcon *icon)
   assert (wlr_drag != NULL);
 
   switch (seat->seat->drag->grab_type) {
-  case WLR_DRAG_GRAB_KEYBOARD:
-    assert (false);
   case WLR_DRAG_GRAB_KEYBOARD_POINTER:;
     struct wlr_cursor *cursor = seat->cursor->cursor;
     icon->x = cursor->x;
@@ -816,6 +803,7 @@ phoc_drag_icon_update_position (PhocDragIcon *icon)
     icon->x = seat->touch_x;
     icon->y = seat->touch_y;
     break;
+  case WLR_DRAG_GRAB_KEYBOARD:
   default:
     g_error ("Invalid drag grab type %d", seat->seat->drag->grab_type);
   }
@@ -997,7 +985,7 @@ seat_add_pointer (PhocSeat                *seat,
 static void
 handle_switch_destroy (struct wl_listener *listener, void *data)
 {
-  struct roots_switch *switch_device =
+  struct phoc_switch *switch_device =
     wl_container_of (listener, switch_device, device_destroy);
   PhocSeat *seat = switch_device->seat;
 
@@ -1013,7 +1001,7 @@ seat_add_switch (PhocSeat                *seat,
                  struct wlr_input_device *device)
 {
   assert (device->type == WLR_INPUT_DEVICE_SWITCH);
-  struct roots_switch *switch_device = g_new0 (struct roots_switch, 1);
+  struct phoc_switch *switch_device = g_new0 (struct phoc_switch, 1);
 
   device->data = switch_device;
   switch_device->device = device;
@@ -1347,6 +1335,13 @@ phoc_seat_has_meta_pressed (PhocSeat *seat)
   return false;
 }
 
+
+/**
+ * phoc_seat_get_focus:
+ * @seat: The seat
+ *
+ * Returns: (nullable)(transfer none): The currently focused view
+ */
 PhocView *
 phoc_seat_get_focus (PhocSeat *seat)
 {
@@ -1425,6 +1420,13 @@ seat_add_view (PhocSeat *seat, PhocView *view)
   return seat_view;
 }
 
+/**
+ * phoc_seat_view_from_view:
+ * @seat: The seat
+ * @view: A view
+ *
+ * Returns: (nullable)(transfer none): The seat view belonging to the given view.
+ */
 PhocSeatView *
 phoc_seat_view_from_view (PhocSeat *seat, PhocView *view)
 {
@@ -1553,7 +1555,7 @@ phoc_seat_set_focus (PhocSeat *seat, PhocView *view)
 
   // Deactivate the old view if it is not focused by some other seat
   if (prev_focus != NULL && !phoc_input_view_has_focus (seat->input, prev_focus)) {
-    view_activate (prev_focus, false);
+    phoc_view_activate (prev_focus, false);
   }
 
   if (view == NULL) {
@@ -1570,7 +1572,7 @@ phoc_seat_set_focus (PhocSeat *seat, PhocView *view)
     return;
   }
 
-  view_activate (view, true);
+  phoc_view_activate (view, true);
   seat->has_focus = true;
 
   // An existing keyboard grab might try to deny setting focus, so cancel it
@@ -1606,9 +1608,9 @@ void
 phoc_seat_set_focus_layer (PhocSeat                    *seat,
                            struct wlr_layer_surface_v1 *layer)
 {
-  PhocServer *server = phoc_server_get_default ();
   if (!layer) {
     if (seat->focused_layer) {
+      PhocOutput *output = PHOC_OUTPUT (seat->focused_layer->output->data);
       seat->focused_layer = NULL;
       if (!wl_list_empty (&seat->views)) {
         // Focus first view
@@ -1618,10 +1620,8 @@ phoc_seat_set_focus_layer (PhocSeat                    *seat,
       } else {
         phoc_seat_set_focus (seat, NULL);
       }
-      PhocOutput *output;
-      wl_list_for_each (output, &server->desktop->outputs, link) {
-        phoc_layer_shell_arrange (output);
-      }
+      phoc_layer_shell_arrange (output);
+      phoc_output_update_shell_reveal (output);
     }
     return;
   }
@@ -1633,7 +1633,7 @@ phoc_seat_set_focus_layer (PhocSeat                    *seat,
   if (seat->has_focus) {
     PhocView *prev_focus = phoc_seat_get_focus (seat);
     wlr_seat_keyboard_clear_focus (seat->seat);
-    view_activate (prev_focus, false);
+    phoc_view_activate (prev_focus, false);
   }
   seat->has_focus = false;
   if (layer->current.layer >= ZWLR_LAYER_SHELL_V1_LAYER_TOP) {
@@ -1650,6 +1650,7 @@ phoc_seat_set_focus_layer (PhocSeat                    *seat,
 
   phoc_cursor_update_focus (seat->cursor);
   phoc_input_method_relay_set_focus (&seat->im_relay, layer->surface);
+  phoc_output_update_shell_reveal (PHOC_OUTPUT (layer->output->data));
 }
 
 void
