@@ -5,7 +5,7 @@
 #include <stdlib.h>
 #include "switch.h"
 
-void roots_switch_handle_toggle(struct roots_switch *switch_device,
+void phoc_switch_handle_toggle(struct phoc_switch *switch_device,
 		struct wlr_event_switch_toggle *event) {
 	g_debug ("Switch %s, type: %d, state: %d",
 		 event->device->name,
