@@ -43,9 +43,14 @@ phoc_xdg_activation_v1_handle_request_activate (struct wl_listener *listener,
   if (view == NULL)
     return;
 
-  g_debug ("Activating view %p via token '%s'", view, token->token);
-  view_activate (view, true);
-  phoc_phosh_private_notify_startup_id (server->desktop->phosh,
-                                        token->token,
-                                        PHOSH_PRIVATE_STARTUP_TRACKER_PROTOCOL_XDG_ACTIVATION);
+  if (phoc_view_is_mapped (view)) {
+     g_debug ("Activating view %p via token '%s'", view, token->token);
+      phoc_view_activate (view, true);
+      phoc_phosh_private_notify_startup_id (server->desktop->phosh,
+                                            token->token,
+                                            PHOSH_PRIVATE_STARTUP_TRACKER_PROTOCOL_XDG_ACTIVATION);
+  } else {
+    g_debug ("Setting view %p via token '%s' as pending activation", view, token->token);
+    phoc_view_set_activation_token (view, token->token);
+  }
 }
